@@ -22,13 +22,13 @@
 
 #define _GNU_SOURCE
 
+#include <math.h>
 #include <stdlib.h>
 #include <stdint.h>
 #include <string.h>
 #include <stdio.h>
 #include <errno.h>
 #include <sys/uio.h>
-#include <ffi.h>
 #include <assert.h>
 #include <fcntl.h>
 #include <unistd.h>
@@ -38,32 +38,25 @@
 
 #include "wayland-util.h"
 #include "wayland-private.h"
+#include "wayland-os.h"
+
+#define DIV_ROUNDUP(n, a) ( ((n) + ((a) - 1)) / (a) )
 
 struct wl_buffer {
 	char data[4096];
-	int head, tail;
+	uint32_t head, tail;
 };
 
 #define MASK(i) ((i) & 4095)
 
-struct wl_closure {
-	int count;
-	const struct wl_message *message;
-	ffi_type *types[20];
-	ffi_cif cif;
-	void *args[20];
-	uint32_t buffer[256];
-	uint32_t *start;
-};
+#define MAX_FDS_OUT	28
+#define CLEN		(CMSG_LEN(MAX_FDS_OUT * sizeof(int32_t)))
 
 struct wl_connection {
 	struct wl_buffer in, out;
 	struct wl_buffer fds_in, fds_out;
 	int fd;
-	void *data;
-	wl_connection_update_func_t update;
-	struct wl_closure receive_closure, send_closure;
-	int write_signalled;
+	int want_flush;
 };
 
 union wl_value {
@@ -77,7 +70,7 @@ union wl_value {
 static void
 wl_buffer_put(struct wl_buffer *b, const void *data, size_t count)
 {
-	int head, size;
+	uint32_t head, size;
 
 	head = MASK(b->head);
 	if (head + count <= sizeof b->data) {
@@ -94,7 +87,7 @@ wl_buffer_put(struct wl_buffer *b, const void *data, size_t count)
 static void
 wl_buffer_put_iov(struct wl_buffer *b, struct iovec *iov, int *count)
 {
-	int head, tail;
+	uint32_t head, tail;
 
 	head = MASK(b->head);
 	tail = MASK(b->tail);
@@ -118,7 +111,7 @@ wl_buffer_put_iov(struct wl_buffer *b, struct iovec *iov, int *count)
 static void
 wl_buffer_get_iov(struct wl_buffer *b, struct iovec *iov, int *count)
 {
-	int head, tail;
+	uint32_t head, tail;
 
 	head = MASK(b->head);
 	tail = MASK(b->tail);
@@ -142,7 +135,7 @@ wl_buffer_get_iov(struct wl_buffer *b, struct iovec *iov, int *count)
 static void
 wl_buffer_copy(struct wl_buffer *b, void *data, size_t count)
 {
-	int tail, size;
+	uint32_t tail, size;
 
 	tail = MASK(b->tail);
 	if (tail + count <= sizeof b->data) {
@@ -154,10 +147,14 @@ wl_buffer_copy(struct wl_buffer *b, void *data, size_t count)
 	}
 }
 
+static uint32_t
+wl_buffer_size(struct wl_buffer *b)
+{
+	return b->head - b->tail;
+}
+
 struct wl_connection *
-wl_connection_create(int fd,
-		     wl_connection_update_func_t update,
-		     void *data)
+wl_connection_create(int fd)
 {
 	struct wl_connection *connection;
 
@@ -166,19 +163,34 @@ wl_connection_create(int fd,
 		return NULL;
 	memset(connection, 0, sizeof *connection);
 	connection->fd = fd;
-	connection->update = update;
-	connection->data = data;
-
-	connection->update(connection,
-			   WL_CONNECTION_READABLE,
-			   connection->data);
 
 	return connection;
 }
 
+static void
+close_fds(struct wl_buffer *buffer, int max)
+{
+	int32_t fds[sizeof(buffer->data) / sizeof(int32_t)], i, count;
+	size_t size;
+
+	size = buffer->head - buffer->tail;
+	if (size == 0)
+		return;
+
+	wl_buffer_copy(buffer, fds, size);
+	count = size / sizeof fds[0];
+	if (max > 0 && max < count)
+		count = max;
+	for (i = 0; i < count; i++)
+		close(fds[i]);
+	buffer->tail += size;
+}
+
 void
 wl_connection_destroy(struct wl_connection *connection)
 {
+	close_fds(&connection->fds_out, -1);
+	close_fds(&connection->fds_in, -1);
 	close(connection->fd);
 	free(connection);
 }
@@ -202,6 +214,9 @@ build_cmsg(struct wl_buffer *buffer, char *data, int *clen)
 	size_t size;
 
 	size = buffer->head - buffer->tail;
+	if (size > MAX_FDS_OUT * sizeof(int32_t))
+		size = MAX_FDS_OUT * sizeof(int32_t);
+
 	if (size > 0) {
 		cmsg = (struct cmsghdr *) data;
 		cmsg->cmsg_level = SOL_SOCKET;
@@ -214,48 +229,53 @@ build_cmsg(struct wl_buffer *buffer, char *data, int *clen)
 	}
 }
 
-static void
-close_fds(struct wl_buffer *buffer)
-{
-	int fds[32], i, count;
-	size_t size;
-
-	size = buffer->head - buffer->tail;
-	if (size == 0)
-		return;
-
-	wl_buffer_copy(buffer, fds, size);
-	count = size / sizeof fds[0];
-	for (i = 0; i < count; i++)
-		close(fds[i]);
-	buffer->tail += size;
-}
-
-static void
+static int
 decode_cmsg(struct wl_buffer *buffer, struct msghdr *msg)
 {
 	struct cmsghdr *cmsg;
-	size_t size;
+	size_t size, max, i;
+	int overflow = 0;
 
 	for (cmsg = CMSG_FIRSTHDR(msg); cmsg != NULL;
 	     cmsg = CMSG_NXTHDR(msg, cmsg)) {
-		if (cmsg->cmsg_level == SOL_SOCKET &&
-		    cmsg->cmsg_type == SCM_RIGHTS) {
-			size = cmsg->cmsg_len - CMSG_LEN(0);
+		if (cmsg->cmsg_level != SOL_SOCKET ||
+		    cmsg->cmsg_type != SCM_RIGHTS)
+			continue;
+
+		size = cmsg->cmsg_len - CMSG_LEN(0);
+		max = sizeof(buffer->data) - wl_buffer_size(buffer);
+		if (size > max || overflow) {
+			overflow = 1;
+			size /= sizeof(int32_t);
+			for (i = 0; i < size; ++i)
+				close(((int*)CMSG_DATA(cmsg))[i]);
+		} else {
 			wl_buffer_put(buffer, CMSG_DATA(cmsg), size);
 		}
 	}
+
+	if (overflow) {
+		errno = EOVERFLOW;
+		return -1;
+	}
+
+	return 0;
 }
 
 int
-wl_connection_data(struct wl_connection *connection, uint32_t mask)
+wl_connection_flush(struct wl_connection *connection)
 {
 	struct iovec iov[2];
 	struct msghdr msg;
-	char cmsg[128];
-	int len, count, clen;
+	char cmsg[CLEN];
+	int len = 0, count, clen;
+	uint32_t tail;
 
-	if (mask & WL_CONNECTION_WRITABLE) {
+	if (!connection->want_flush)
+		return 0;
+
+	tail = connection->out.tail;
+	while (connection->out.head - connection->out.tail > 0) {
 		wl_buffer_get_iov(&connection->out, iov, &count);
 
 		build_cmsg(&connection->fds_out, cmsg, &clen);
@@ -269,91 +289,93 @@ wl_connection_data(struct wl_connection *connection, uint32_t mask)
 		msg.msg_flags = 0;
 
 		do {
-			len = sendmsg(connection->fd, &msg, MSG_NOSIGNAL);
-		} while (len < 0 && errno == EINTR);
+			len = sendmsg(connection->fd, &msg,
+				      MSG_NOSIGNAL | MSG_DONTWAIT);
+		} while (len == -1 && errno == EINTR);
 
-		if (len == -1 && errno == EPIPE) {
+		if (len == -1)
 			return -1;
-		} else if (len < 0) {
-			fprintf(stderr,
-				"write error for connection %p, fd %d: %m\n",
-				connection, connection->fd);
-			return -1;
-		}
 
-		close_fds(&connection->fds_out);
+		close_fds(&connection->fds_out, MAX_FDS_OUT);
 
 		connection->out.tail += len;
-		if (connection->out.tail == connection->out.head &&
-		    connection->write_signalled) {
-			connection->update(connection,
-					   WL_CONNECTION_READABLE,
-					   connection->data);
-			connection->write_signalled = 0;
-		}
 	}
 
-	if (mask & WL_CONNECTION_READABLE) {
-		wl_buffer_put_iov(&connection->in, iov, &count);
+	connection->want_flush = 0;
 
-		msg.msg_name = NULL;
-		msg.msg_namelen = 0;
-		msg.msg_iov = iov;
-		msg.msg_iovlen = count;
-		msg.msg_control = cmsg;
-		msg.msg_controllen = sizeof cmsg;
-		msg.msg_flags = 0;
+	return connection->out.head - tail;
+}
 
-		do {
-			len = recvmsg(connection->fd, &msg, MSG_CMSG_CLOEXEC);
-		} while (len < 0 && errno == EINTR);
+int
+wl_connection_read(struct wl_connection *connection)
+{
+	struct iovec iov[2];
+	struct msghdr msg;
+	char cmsg[CLEN];
+	int len, count, ret;
 
-		if (len < 0) {
-			fprintf(stderr,
-				"read error from connection %p: %m (%d)\n",
-				connection, errno);
-			return -1;
-		} else if (len == 0) {
-			/* FIXME: Handle this better? */
-			return -1;
-		}
+	if (wl_buffer_size(&connection->in) >= sizeof(connection->in.data)) {
+		errno = EOVERFLOW;
+		return -1;
+	}
 
-		decode_cmsg(&connection->fds_in, &msg);
+	wl_buffer_put_iov(&connection->in, iov, &count);
 
-		connection->in.head += len;
-	}	
+	msg.msg_name = NULL;
+	msg.msg_namelen = 0;
+	msg.msg_iov = iov;
+	msg.msg_iovlen = count;
+	msg.msg_control = cmsg;
+	msg.msg_controllen = sizeof cmsg;
+	msg.msg_flags = 0;
+
+	do {
+		len = wl_os_recvmsg_cloexec(connection->fd, &msg, 0);
+	} while (len < 0 && errno == EINTR);
+
+	if (len <= 0)
+		return len;
+
+	ret = decode_cmsg(&connection->fds_in, &msg);
+	if (ret)
+		return -1;
+
+	connection->in.head += len;
 
 	return connection->in.head - connection->in.tail;
 }
 
-void
+int
 wl_connection_write(struct wl_connection *connection,
 		    const void *data, size_t count)
 {
 	if (connection->out.head - connection->out.tail +
-	    count > ARRAY_LENGTH(connection->out.data))
-		wl_connection_data(connection, WL_CONNECTION_WRITABLE);
+	    count > ARRAY_LENGTH(connection->out.data)) {
+		connection->want_flush = 1;
+		if (wl_connection_flush(connection) < 0)
+			return -1;
+	}
 
 	wl_buffer_put(&connection->out, data, count);
+	connection->want_flush = 1;
 
-	if (!connection->write_signalled) {
-		connection->update(connection,
-				   WL_CONNECTION_READABLE |
-				   WL_CONNECTION_WRITABLE,
-				   connection->data);
-		connection->write_signalled = 1;
-	}
+	return 0;
 }
 
-static void
+int
 wl_connection_queue(struct wl_connection *connection,
 		    const void *data, size_t count)
 {
 	if (connection->out.head - connection->out.tail +
-	    count > ARRAY_LENGTH(connection->out.data))
-		wl_connection_data(connection, WL_CONNECTION_WRITABLE);
+	    count > ARRAY_LENGTH(connection->out.data)) {
+		connection->want_flush = 1;
+		if (wl_connection_flush(connection) < 0)
+			return -1;
+	}
 
 	wl_buffer_put(&connection->out, data, count);
+
+	return 0;
 }
 
 static int
@@ -366,6 +388,7 @@ wl_message_size_extra(const struct wl_message *message)
 		switch (message->signature[i]) {
 		case 's':
 		case 'o':
+		case 'n':
 			extra += sizeof (void *);
 			break;
 		case 'a':
@@ -382,30 +405,87 @@ wl_message_size_extra(const struct wl_message *message)
 	return extra;
 }
 
-struct wl_closure *
-wl_connection_vmarshal(struct wl_connection *connection,
-		       struct wl_object *sender,
-		       uint32_t opcode, va_list ap,
-		       const struct wl_message *message)
+static int
+wl_connection_put_fd(struct wl_connection *connection, int32_t fd)
 {
-	struct wl_closure *closure = &connection->send_closure;
+	if (wl_buffer_size(&connection->fds_out) == MAX_FDS_OUT * sizeof fd) {
+		connection->want_flush = 1;
+		if (wl_connection_flush(connection) < 0)
+			return -1;
+	}
+
+	wl_buffer_put(&connection->fds_out, &fd, sizeof fd);
+
+	return 0;
+}
+
+const char *
+get_next_argument(const char *signature, struct argument_details *details)
+{
+	if (*signature == '?') {
+		details->nullable = 1;
+		signature++;
+	} else
+		details->nullable = 0;
+
+	details->type = *signature;
+	return signature + 1;
+}
+
+int
+arg_count_for_signature(const char *signature)
+{
+	int count = 0;
+	while (*signature) {
+		if (*signature != '?')
+			count++;
+		signature++;
+	}
+	return count;
+}
+
+struct wl_closure *
+wl_closure_vmarshal(struct wl_object *sender,
+		    uint32_t opcode, va_list ap,
+		    const struct wl_message *message)
+{
+	struct wl_closure *closure;
 	struct wl_object **objectp, *object;
-	uint32_t length, *p, *start, size, *end;
+	uint32_t length, aligned, *p, *start, size, *end;
 	int dup_fd;
 	struct wl_array **arrayp, *array;
 	const char **sp, *s;
+	const char *signature = message->signature;
+	struct argument_details arg;
 	char *extra;
 	int i, count, fd, extra_size, *fd_ptr;
 
+	/* FIXME: Match old fixed allocation for now */
+	closure = malloc(sizeof *closure + 1024);
+	if (closure == NULL)
+		return NULL;
+
 	extra_size = wl_message_size_extra(message);
-	count = strlen(message->signature) + 2;
+	count = arg_count_for_signature(signature) + 2;
 	extra = (char *) closure->buffer;
 	start = &closure->buffer[DIV_ROUNDUP(extra_size, sizeof *p)];
-	end = &closure->buffer[ARRAY_LENGTH(closure->buffer)];
+	end = &closure->buffer[256];
 	p = &start[2];
 
+	closure->types[0] = &ffi_type_pointer;
+	closure->types[1] = &ffi_type_pointer;
+
 	for (i = 2; i < count; i++) {
-		switch (message->signature[i - 2]) {
+		signature = get_next_argument(signature, &arg);
+
+		switch (arg.type) {
+		case 'f':
+			closure->types[i] = &ffi_type_sint32;
+			closure->args[i] = p;
+			if (end - p < 1)
+				goto err;
+			*p++ = va_arg(ap, wl_fixed_t);
+			break;
 		case 'u':
 			closure->types[i] = &ffi_type_uint32;
 			closure->args[i] = p;
@@ -427,18 +507,24 @@ wl_connection_vmarshal(struct wl_connection *connection,
 			extra += sizeof *sp;
 
 			s = va_arg(ap, const char *);
+
+			if (!arg.nullable && s == NULL)
+				goto err_null;
+
 			length = s ? strlen(s) + 1: 0;
-			if (end - p < DIV_ROUNDUP(length, sizeof *p) + 1)
+			aligned = (length + 3) & ~3;
+			if (p + aligned / sizeof *p + 1 > end)
 				goto err;
 			*p++ = length;
 
-			if (length > 0)
+			if (length > 0) {
+				memcpy(p, s, length);
 				*sp = (const char *) p;
-			else
+			} else
 				*sp = NULL;
 
-			memcpy(p, s, length);
-			p += DIV_ROUNDUP(length, sizeof *p);
+			memset((char *) p + length, 0, aligned - length);
+			p += aligned / sizeof *p;
 			break;
 		case 'o':
 			closure->types[i] = &ffi_type_pointer;
@@ -447,6 +533,10 @@ wl_connection_vmarshal(struct wl_connection *connection,
 			extra += sizeof *objectp;
 
 			object = va_arg(ap, struct wl_object *);
+
+			if (!arg.nullable && object == NULL)
+				goto err_null;
+
 			*objectp = object;
 			if (end - p < 1)
 				goto err;
@@ -459,6 +549,10 @@ wl_connection_vmarshal(struct wl_connection *connection,
 			object = va_arg(ap, struct wl_object *);
 			if (end - p < 1)
 				goto err;
+
+			if (!arg.nullable && object == NULL)
+				goto err_null;
+
 			*p++ = object ? object->id : 0;
 			break;
 
@@ -472,13 +566,17 @@ wl_connection_vmarshal(struct wl_connection *connection,
 			extra += sizeof **arrayp;
 
 			array = va_arg(ap, struct wl_array *);
+
+			if (!arg.nullable && array == NULL)
+				goto err_null;
+
 			if (array == NULL || array->size == 0) {
 				if (end - p < 1)
 					goto err;
 				*p++ = 0;
 				break;
 			}
-			if (end - p < DIV_ROUNDUP(array->size, sizeof *p) + 1)
+			if (p + DIV_ROUNDUP(array->size, sizeof *p) + 1 > end)
 				goto err;
 			*p++ = array->size;
 			memcpy(p, array->data, array->size);
@@ -497,18 +595,16 @@ wl_connection_vmarshal(struct wl_connection *connection,
 			extra += sizeof *fd_ptr;
 
 			fd = va_arg(ap, int);
-			dup_fd = fcntl(fd, F_DUPFD_CLOEXEC, 0);
+			dup_fd = wl_os_dupfd_cloexec(fd, 0);
 			if (dup_fd < 0) {
 				fprintf(stderr, "dup failed: %m");
 				abort();
 			}
 			*fd_ptr = dup_fd;
-			wl_buffer_put(&connection->fds_out,
-				      &dup_fd, sizeof dup_fd);
 			break;
 		default:
 			fprintf(stderr, "unhandled format code: '%c'\n",
-				message->signature[i - 2]);
+				arg.type);
 			assert(0);
 			break;
 		}
@@ -522,12 +618,26 @@ wl_connection_vmarshal(struct wl_connection *connection,
 	closure->message = message;
 	closure->count = count;
 
+	ffi_prep_cif(&closure->cif, FFI_DEFAULT_ABI,
+		     closure->count, &ffi_type_void, closure->types);
+
 	return closure;
 
 err:
-	printf("request too big to marshal, maximum size is %lu\n",
+	printf("request too big to marshal, maximum size is %zu\n",
 	       sizeof closure->buffer);
 	errno = ENOMEM;
+	free(closure);
+
+	return NULL;
+
+err_null:
+	free(closure);
+	wl_log("error marshalling arguments for %s:%i.%s (signature %s): "
+	       "null value passed for arg %i\n",
+	       sender->interface->name, sender->id, message->name,
+	       message->signature, i);
+	errno = EINVAL;
 	return NULL;
 }
 
@@ -537,15 +647,16 @@ wl_connection_demarshal(struct wl_connection *connection,
 			struct wl_map *objects,
 			const struct wl_message *message)
 {
-	uint32_t *p, *next, *end, length;
+	uint32_t *p, *next, *end, length, **id;
 	int *fd;
 	char *extra, **s;
-	int i, count, extra_space;
-	struct wl_object **object;
+	unsigned int i, count, extra_space;
+	const char *signature = message->signature;
+	struct argument_details arg;
 	struct wl_array **array;
-	struct wl_closure *closure = &connection->receive_closure;
+	struct wl_closure *closure;
 
-	count = strlen(message->signature) + 2;
+	count = arg_count_for_signature(signature) + 2;
 	if (count > ARRAY_LENGTH(closure->types)) {
 		printf("too many args (%d)\n", count);
 		errno = EINVAL;
@@ -554,23 +665,22 @@ wl_connection_demarshal(struct wl_connection *connection,
 	}
 
 	extra_space = wl_message_size_extra(message);
-	if (sizeof closure->buffer < size + extra_space) {
-		printf("request too big to demarshal, maximum %lu actual %d\n",
-		       sizeof closure->buffer, size + extra_space);
-		errno = ENOMEM;
-		wl_connection_consume(connection, size);
+	closure = malloc(sizeof *closure + 8 + size + extra_space);
+	if (closure == NULL)
 		return NULL;
-	}
 
 	closure->message = message;
 	closure->types[0] = &ffi_type_pointer;
 	closure->types[1] = &ffi_type_pointer;
+	closure->start = closure->buffer;
 
 	wl_connection_copy(connection, closure->buffer, size);
 	p = &closure->buffer[2];
 	end = (uint32_t *) ((char *) p + size);
 	extra = (char *) end;
 	for (i = 2; i < count; i++) {
+		signature = get_next_argument(signature, &arg);
+
 		if (p + 1 > end) {
 			printf("message too short, "
 			       "object (%d), message %s(%s)\n",
@@ -579,7 +689,7 @@ wl_connection_demarshal(struct wl_connection *connection,
 			goto err;
 		}
 
-		switch (message->signature[i - 2]) {
+		switch (arg.type) {
 		case 'u':
 			closure->types[i] = &ffi_type_uint32;
 			closure->args[i] = p++;
@@ -588,6 +698,10 @@ wl_connection_demarshal(struct wl_connection *connection,
 			closure->types[i] = &ffi_type_sint32;
 			closure->args[i] = p++;
 			break;
+		case 'f':
+			closure->types[i] = &ffi_type_sint32;
+			closure->args[i] = p++;
+			break;
 		case 's':
 			closure->types[i] = &ffi_type_pointer;
 			length = *p++;
@@ -622,29 +736,15 @@ wl_connection_demarshal(struct wl_connection *connection,
 			break;
 		case 'o':
 			closure->types[i] = &ffi_type_pointer;
-			object = (struct wl_object **) extra;
-			extra += sizeof *object;
-			closure->args[i] = object;
+			id = (uint32_t **) extra;
+			extra += sizeof *id;
+			closure->args[i] = id;
+			*id = p;
 
-			*object = wl_map_lookup(objects, *p);
-			if (*object == WL_ZOMBIE_OBJECT) {
-				/* references object we've already
-				 * destroyed client side */
-				*object = NULL;
-			} else if (*object == NULL && *p != 0) {
-				printf("unknown object (%d), message %s(%s)\n",
-				       *p, message->name, message->signature);
-				*object = NULL;
-				errno = EINVAL;
-				goto err;
-			}
-
-			if (*object != NULL && message->types[i-2] != NULL &&
-			    (*object)->interface != message->types[i-2]) {
-				printf("invalid object (%d), type (%s), "
-					"message %s(%s)\n",
-				       *p, (*object)->interface->name,
-				       message->name, message->signature);
+			if (**id == 0 && !arg.nullable) {
+				printf("NULL object received on non-nullable "
+				       "type, message %s(%s)\n", message->name,
+				       message->signature);
 				errno = EINVAL;
 				goto err;
 			}
@@ -652,16 +752,28 @@ wl_connection_demarshal(struct wl_connection *connection,
 			p++;
 			break;
 		case 'n':
-			closure->types[i] = &ffi_type_uint32;
-			closure->args[i] = p;
-			object = wl_map_lookup(objects, *p);
-			if (object != NULL) {
-				printf("not a new object (%d), "
+			closure->types[i] = &ffi_type_pointer;
+			id = (uint32_t **) extra;
+			extra += sizeof *id;
+			closure->args[i] = id;
+			*id = p;
+
+			if (**id == 0 && !arg.nullable) {
+				printf("NULL new ID received on non-nullable "
+				       "type, message %s(%s)\n", message->name,
+				       message->signature);
+				errno = EINVAL;
+				goto err;
+			}
+
+			if (wl_map_reserve_new(objects, *p) < 0) {
+				printf("not a valid new object id (%d), "
 				       "message %s(%s)\n",
 				       *p, message->name, message->signature);
 				errno = EINVAL;
 				goto err;
 			}
+
 			p++;
 			break;
 		case 'a':
@@ -707,6 +819,7 @@ wl_connection_demarshal(struct wl_connection *connection,
 	}
 
 	closure->count = i;
+
 	ffi_prep_cif(&closure->cif, FFI_DEFAULT_ABI,
 		     closure->count, &ffi_type_void, closure->types);
 
@@ -722,6 +835,67 @@ wl_connection_demarshal(struct wl_connection *connection,
 	return NULL;
 }
 
+static int
+interface_equal(const struct wl_interface *a, const struct wl_interface *b)
+{
+	/* In most cases the pointer equality test is sufficient.
+	 * However, in some cases, depending on how things are split
+	 * across shared objects, we can end up with multiple
+	 * instances of the interface metadata constants.  So if the
+	 * pointers match, the interfaces are equal, if they don't
+	 * match we have to compare the interface names. */
+
+	return a == b || strcmp(a->name, b->name) == 0;
+}
+
+int
+wl_closure_lookup_objects(struct wl_closure *closure, struct wl_map *objects)
+{
+	struct wl_object **object;
+	const struct wl_message *message;
+	const char *signature;
+	struct argument_details arg;
+	int i, count;
+	uint32_t id;
+
+	message = closure->message;
+	signature = message->signature;
+	count = arg_count_for_signature(signature) + 2;
+	for (i = 2; i < count; i++) {
+		signature = get_next_argument(signature, &arg);
+		switch (arg.type) {
+		case 'o':
+			id = **(uint32_t **) closure->args[i];
+			object = closure->args[i];
+			*object = wl_map_lookup(objects, id);
+			if (*object == WL_ZOMBIE_OBJECT) {
+				/* references object we've already
+				 * destroyed client side */
+				*object = NULL;
+			} else if (*object == NULL && id != 0) {
+				printf("unknown object (%u), message %s(%s)\n",
+				       id, message->name, message->signature);
+				*object = NULL;
+				errno = EINVAL;
+				return -1;
+			}
+
+			if (*object != NULL && message->types[i-2] != NULL &&
+			    !interface_equal((*object)->interface,
+					     message->types[i-2])) {
+				printf("invalid object (%u), type (%s), "
+				       "message %s(%s)\n",
+				       id, (*object)->interface->name,
+				       message->name, message->signature);
+				errno = EINVAL;
+				return -1;
+			}
+		}
+	}
+
+	return 0;
+}
+
 void
 wl_closure_invoke(struct wl_closure *closure,
 		  struct wl_object *target, void (*func)(void), void *data)
@@ -734,29 +908,67 @@ wl_closure_invoke(struct wl_closure *closure,
 	ffi_call(&closure->cif, func, &result, closure->args);
 }
 
-void
+static int
+copy_fds_to_connection(struct wl_closure *closure,
+		       struct wl_connection *connection)
+{
+	const struct wl_message *message = closure->message;
+	uint32_t i, count;
+	struct argument_details arg;
+	const char *signature = message->signature;
+	int *fd;
+
+	count = arg_count_for_signature(signature) + 2;
+	for (i = 2; i < count; i++) {
+		signature = get_next_argument(signature, &arg);
+		if (arg.type != 'h')
+			continue;
+
+		fd = closure->args[i];
+		if (wl_connection_put_fd(connection, *fd)) {
+			fprintf(stderr, "request could not be marshaled: "
+				"can't send file descriptor");
+			return -1;
+		}
+	}
+
+	return 0;
+}
+
+int
 wl_closure_send(struct wl_closure *closure, struct wl_connection *connection)
 {
 	uint32_t size;
 
+	if (copy_fds_to_connection(closure, connection))
+		return -1;
+
 	size = closure->start[1] >> 16;
-	wl_connection_write(connection, closure->start, size);
+
+	return wl_connection_write(connection, closure->start, size);
 }
 
-void
+int
 wl_closure_queue(struct wl_closure *closure, struct wl_connection *connection)
 {
 	uint32_t size;
 
+	if (copy_fds_to_connection(closure, connection))
+		return -1;
+
 	size = closure->start[1] >> 16;
-	wl_connection_queue(connection, closure->start, size);
+
+	return wl_connection_queue(connection, closure->start, size);
 }
 
 void
 wl_closure_print(struct wl_closure *closure, struct wl_object *target, int send)
 {
 	union wl_value *value;
+	int32_t si;
 	int i;
+	struct argument_details arg;
+	const char *signature = closure->message->signature;
 	struct timespec tp;
 	unsigned int time;
 
@@ -770,16 +982,22 @@ wl_closure_print(struct wl_closure *closure, struct wl_object *target, int send)
 		closure->message->name);
 
 	for (i = 2; i < closure->count; i++) {
+		signature = get_next_argument(signature, &arg);
 		if (i > 2)
 			fprintf(stderr, ", ");
 
 		value = closure->args[i];
-		switch (closure->message->signature[i - 2]) {
+		switch (arg.type) {
 		case 'u':
 			fprintf(stderr, "%u", value->uint32);
 			break;
 		case 'i':
-			fprintf(stderr, "%d", value->uint32);
+			si = (int32_t) value->uint32;
+			fprintf(stderr, "%d", si);
+			break;
+		case 'f':
+			si = (int32_t) value->uint32;
+			fprintf(stderr, "%f", wl_fixed_to_double(si));
 			break;
 		case 's':
 			fprintf(stderr, "\"%s\"", value->string);
@@ -793,7 +1011,16 @@ wl_closure_print(struct wl_closure *closure, struct wl_object *target, int send)
 				fprintf(stderr, "nil");
 			break;
 		case 'n':
-			fprintf(stderr, "new id %u", value->uint32);
+			fprintf(stderr, "new id %s@",
+				(closure->message->types[i - 2]) ?
+				 closure->message->types[i - 2]->name :
+				  "[unknown]");
+			if (send && value->new_id != 0)
+				fprintf(stderr, "%u", value->new_id);
+			else if (!send && value->object != NULL)
+				fprintf(stderr, "%u", value->object->id);
+			else
+				fprintf(stderr, "nil");
 			break;
 		case 'a':
 			fprintf(stderr, "array");
@@ -810,4 +1037,5 @@ wl_closure_print(struct wl_closure *closure, struct wl_object *target, int send)
 void
 wl_closure_destroy(struct wl_closure *closure)
 {
+	free(closure);
 }
