@@ -22,6 +22,7 @@
  */
 
 #include <stdio.h>
+#include <stdarg.h>
 #include <string.h>
 #include <errno.h>
 #include <ctype.h>
@@ -57,6 +58,7 @@ struct interface {
 	char *name;
 	char *uppercase_name;
 	int version;
+	int since;
 	struct wl_list request_list;
 	struct wl_list event_list;
 	struct wl_list enumeration_list;
@@ -73,6 +75,7 @@ struct message {
 	int type_index;
 	int all_null;
 	int destructor;
+	int since;
 	struct description *description;
 };
 
@@ -80,6 +83,7 @@ enum arg_type {
 	NEW_ID,
 	INT,
 	UNSIGNED,
+	FIXED,
 	STRING,
 	OBJECT,
 	ARRAY,
@@ -89,6 +93,7 @@ enum arg_type {
 struct arg {
 	char *name;
 	enum arg_type type;
+	int nullable;
 	char *interface_name;
 	struct wl_list link;
 	char *summary;
@@ -119,9 +124,21 @@ struct parse_context {
 	struct enumeration *enumeration;
 	struct description *description;
 	char character_data[8192];
-	int character_data_length;
+	unsigned int character_data_length;
 };
 
+static int
+list_length(struct wl_list *list)
+{
+	struct wl_list *l;
+	int i;
+
+	for (i = 0, l = list->next; l != list; i++, l = l->next)
+		;
+
+	return i;
+}
+
 static char *
 uppercase_dup(const char *src)
 {
@@ -136,39 +153,84 @@ uppercase_dup(const char *src)
 	return u;
 }
 
-static void
-desc_dump(char *src, int startcol)
+static const char *indent(int n)
 {
-	int i, j = 0, col = startcol, line = 0, len;
+	const char *whitespace[] = {
+		"\t\t\t\t\t\t\t\t\t\t\t\t",
+		"\t\t\t\t\t\t\t\t\t\t\t\t ",
+		"\t\t\t\t\t\t\t\t\t\t\t\t  ",
+		"\t\t\t\t\t\t\t\t\t\t\t\t   ",
+		"\t\t\t\t\t\t\t\t\t\t\t\t    ",
+		"\t\t\t\t\t\t\t\t\t\t\t\t     ",
+		"\t\t\t\t\t\t\t\t\t\t\t\t      ",
+		"\t\t\t\t\t\t\t\t\t\t\t\t       "
+	};
 
-	/* Strip leading space */
-	for (i = 0; isspace(src[i]); i++)
-		;
+	return whitespace[n % 8] + 12 - n / 8;
+}
 
-	for (; src[i]; i++) {
-		/* Collapse multiple spaces into 1 */
-		if (isspace(src[i]) && isspace(src[i + 1]))
-		    continue;
+static void
+desc_dump(char *desc, const char *fmt, ...) __attribute__((format(printf,2,3)));
 
-		if (isspace(src[i]))
-			src[i] = ' ';
+static void
+desc_dump(char *desc, const char *fmt, ...)
+{
+	va_list ap;
+	char buf[128], hang;
+	int col, i, j, k, startcol, newlines;
 
-		if (col > 72 && isspace(src[i])) {
-			if (src[i+1]) {
-				putchar('\n');
-				for (j = 0; j < startcol; j++)
-					putchar(' ');
-				putchar(' ');
-				putchar('*');
-				putchar(' ');
-			}
-			line++;
-			col = startcol;
-		} else {
-			putchar(src[i]);
+	va_start(ap, fmt);
+	vsnprintf(buf, sizeof buf, fmt, ap);
+	va_end(ap);
+
+	for (i = 0, col = 0; buf[i] != '*'; i++) {
+		if (buf[i] == '\t')
+			col = (col + 8) & ~7;
+		else
 			col++;
-		}
 	}
+
+	printf("%s", buf);
+
+	if (!desc) {
+		printf("(none)\n");
+		return;
+	}
+
+	startcol = col;
+	col += strlen(&buf[i]);
+	if (col - startcol > 2)
+		hang = '\t';
+	else
+		hang = ' ';
+
+	for (i = 0; desc[i]; ) {
+		k = i;
+		newlines = 0;
+		while (desc[i] && isspace(desc[i])) {
+			if (desc[i] == '\n')
+				newlines++;
+			i++;
+		}
+		if (!desc[i])
+			break;
+
+		j = i;
+		while (desc[i] && !isspace(desc[i]))
+			i++;
+
+		if (newlines > 1)
+			printf("\n%s*", indent(startcol));
+		if (newlines > 1 || col + i - j > 72) {
+			printf("\n%s*%c", indent(startcol), hang);
+			col = startcol;
+		}
+
+		if (col > startcol && k > 0)
+			col += printf(" ");
+		col += printf("%.*s", i - j, &desc[j]);
+	}
+	putchar('\n');
 }
 
 static void
@@ -179,6 +241,21 @@ fail(struct parse_context *ctx, const char *msg)
 	exit(EXIT_FAILURE);
 }
 
+static int
+is_nullable_type(struct arg *arg)
+{
+	switch (arg->type) {
+	/* Strings, objects, and arrays are possibly nullable */
+	case STRING:
+	case OBJECT:
+	case NEW_ID:
+	case ARRAY:
+		return 1;
+	default:
+		return 0;
+	}
+}
+
 static void
 start_element(void *data, const char *element_name, const char **atts)
 {
@@ -189,7 +266,9 @@ start_element(void *data, const char *element_name, const char **atts)
 	struct enumeration *enumeration;
 	struct entry *entry;
 	struct description *description;
-	const char *name, *type, *interface_name, *value, *summary;
+	const char *name, *type, *interface_name, *value, *summary, *since;
+	const char *allow_null;
+	char *end;
 	int i, version;
 
 	name = NULL;
@@ -199,6 +278,8 @@ start_element(void *data, const char *element_name, const char **atts)
 	value = NULL;
 	summary = NULL;
 	description = NULL;
+	since = NULL;
+	allow_null = NULL;
 	for (i = 0; atts[i]; i += 2) {
 		if (strcmp(atts[i], "name") == 0)
 			name = atts[i + 1];
@@ -212,6 +293,10 @@ start_element(void *data, const char *element_name, const char **atts)
 			interface_name = atts[i + 1];
 		if (strcmp(atts[i], "summary") == 0)
 			summary = atts[i + 1];
+		if (strcmp(atts[i], "since") == 0)
+			since = atts[i + 1];
+		if (strcmp(atts[i], "allow-null") == 0)
+			allow_null = atts[i + 1];
 	}
 
 	ctx->character_data_length = 0;
@@ -236,6 +321,7 @@ start_element(void *data, const char *element_name, const char **atts)
 		interface->uppercase_name = uppercase_dup(name);
 		interface->version = version;
 		interface->description = NULL;
+		interface->since = 1;
 		wl_list_init(&interface->request_list);
 		wl_list_init(&interface->event_list);
 		wl_list_init(&interface->enumeration_list);
@@ -266,11 +352,25 @@ start_element(void *data, const char *element_name, const char **atts)
 		else
 			message->destructor = 0;
 
+		if (since != NULL) {
+			version = strtol(since, &end, 0);
+			if (errno == EINVAL || end == since || *end != '\0')
+				fail(ctx, "invalid integer\n");
+			if (version <= ctx->interface->since)
+				fail(ctx, "since version not increasing\n");
+			ctx->interface->since = version;
+		}
+
+		message->since = ctx->interface->since;
+
 		if (strcmp(name, "destroy") == 0 && !message->destructor)
 			fail(ctx, "destroy request should be destructor type");
 
 		ctx->message = message;
 	} else if (strcmp(element_name, "arg") == 0) {
+		if (name == NULL)
+			fail(ctx, "no argument name given");
+
 		arg = malloc(sizeof *arg);
 		arg->name = strdup(name);
 
@@ -278,6 +378,8 @@ start_element(void *data, const char *element_name, const char **atts)
 			arg->type = INT;
 		else if (strcmp(type, "uint") == 0)
 			arg->type = UNSIGNED;
+		else if (strcmp(type, "fixed") == 0)
+			arg->type = FIXED;
 		else if (strcmp(type, "string") == 0)
 			arg->type = STRING;
 		else if (strcmp(type, "array") == 0)
@@ -295,9 +397,10 @@ start_element(void *data, const char *element_name, const char **atts)
 		switch (arg->type) {
 		case NEW_ID:
 		case OBJECT:
-			if (interface_name == NULL)
-				fail(ctx, "no interface name given");
-			arg->interface_name = strdup(interface_name);
+			if (interface_name)
+				arg->interface_name = strdup(interface_name);
+			else
+				arg->interface_name = NULL;
 			break;
 		default:
 			if (interface_name != NULL)
@@ -305,6 +408,16 @@ start_element(void *data, const char *element_name, const char **atts)
 			break;
 		}
 
+		if (allow_null == NULL || strcmp(allow_null, "false") == 0)
+			arg->nullable = 0;
+		else if (strcmp(allow_null, "true") == 0)
+			arg->nullable = 1;
+		else
+			fail(ctx, "invalid value for allow-null attribute");
+
+		if (allow_null != NULL && !is_nullable_type(arg))
+			fail(ctx, "allow-null is only valid for objects, strings, and arrays");
+
 		arg->summary = NULL;
 		if (summary)
 			arg->summary = strdup(summary);
@@ -344,10 +457,7 @@ start_element(void *data, const char *element_name, const char **atts)
 			fail(ctx, "description without summary");
 
 		description = malloc(sizeof *description);
-		if (summary)
-			description->summary = strdup(summary);
-		else
-			description->summary = NULL;
+		description->summary = strdup(summary);
 
 		if (ctx->message)
 			ctx->message->description = description;
@@ -371,10 +481,9 @@ end_element(void *data, const XML_Char *name)
 			strndup(ctx->character_data,
 				ctx->character_data_length);
 	} else if (strcmp(name, "description") == 0) {
-		char *text = strndup(ctx->character_data,
-				     ctx->character_data_length);
-		if (text)
-			ctx->description->text = strdup(text);
+		ctx->description->text =
+			strndup(ctx->character_data,
+				ctx->character_data_length);
 		ctx->description = NULL;
 	} else if (strcmp(name, "request") == 0 ||
 		   strcmp(name, "event") == 0) {
@@ -428,6 +537,9 @@ emit_type(struct arg *a)
 	case UNSIGNED:
 		printf("uint32_t ");
 		break;
+	case FIXED:
+		printf("wl_fixed_t ");
+		break;
 	case STRING:
 		printf("const char *");
 		break;
@@ -447,10 +559,6 @@ emit_stubs(struct wl_list *message_list, struct interface *interface)
 	struct arg *a, *ret;
 	int has_destructor, has_destroy;
 
-	/* We provide a hand written functions for the display object */
-	if (strcmp(interface->name, "wl_display") == 0)
-		return;
-
 	printf("static inline void\n"
 	       "%s_set_user_data(struct %s *%s, void *user_data)\n"
 	       "{\n"
@@ -483,7 +591,7 @@ emit_stubs(struct wl_list *message_list, struct interface *interface)
 		exit(EXIT_FAILURE);
 	}
 
-	if (!has_destructor)
+	if (!has_destructor && strcmp(interface->name, "wl_display") != 0)
 		printf("static inline void\n"
 		       "%s_destroy(struct %s *%s)\n"
 		       "{\n"
@@ -503,7 +611,9 @@ emit_stubs(struct wl_list *message_list, struct interface *interface)
 				ret = a;
 		}
 
-		if (ret)
+		if (ret && ret->interface_name == NULL)
+			printf("static inline void *\n");
+		else if (ret)
 			printf("static inline struct %s *\n",
 			       ret->interface_name);
 		else
@@ -514,7 +624,11 @@ emit_stubs(struct wl_list *message_list, struct interface *interface)
 		       interface->name, interface->name);
 
 		wl_list_for_each(a, &m->arg_list, link) {
-			if (a->type == NEW_ID)
+			if (a->type == NEW_ID && a->interface_name == NULL) {
+				printf(", const struct wl_interface *interface"
+				       ", uint32_t version");
+				continue;
+			} else if (a->type == NEW_ID)
 				continue;
 			printf(", ");
 			emit_type(a);
@@ -523,17 +637,21 @@ emit_stubs(struct wl_list *message_list, struct interface *interface)
 
 		printf(")\n"
 		       "{\n");
-		if (ret)
+		if (ret) {
 			printf("\tstruct wl_proxy *%s;\n\n"
 			       "\t%s = wl_proxy_create("
-			       "(struct wl_proxy *) %s,\n"
-			       "\t\t\t     &%s_interface);\n"
-			       "\tif (!%s)\n"
+			       "(struct wl_proxy *) %s,\n",
+			       ret->name, ret->name, interface->name);
+			if (ret->interface_name == NULL)
+				printf("\t\t\t     interface);\n");
+			else
+				printf("\t\t\t     &%s_interface);\n",
+				       ret->interface_name);
+
+			printf("\tif (!%s)\n"
 			       "\t\treturn NULL;\n\n",
-			       ret->name,
-			       ret->name,
-			       interface->name, ret->interface_name,
 			       ret->name);
+		}
 
 		printf("\twl_proxy_marshal((struct wl_proxy *) %s,\n"
 		       "\t\t\t %s_%s",
@@ -542,8 +660,10 @@ emit_stubs(struct wl_list *message_list, struct interface *interface)
 		       m->uppercase_name);
 
 		wl_list_for_each(a, &m->arg_list, link) {
+			if (a->type == NEW_ID && a->interface_name == NULL)
+				printf(", interface->name, version");
 			printf(", ");
-				printf("%s", a->name);
+			printf("%s", a->name);
 		}
 		printf(");\n");
 
@@ -552,7 +672,9 @@ emit_stubs(struct wl_list *message_list, struct interface *interface)
 			       "(struct wl_proxy *) %s);\n",
 			       interface->name);
 
-		if (ret)
+		if (ret && ret->interface_name == NULL)
+			printf("\n\treturn (void *) %s;\n", ret->name);
+		else if (ret)
 			printf("\n\treturn (struct %s *) %s;\n",
 			       ret->interface_name, ret->name);
 
@@ -560,20 +682,45 @@ emit_stubs(struct wl_list *message_list, struct interface *interface)
 	}
 }
 
-static const char *indent(int n)
+static void
+emit_event_wrappers(struct wl_list *message_list, struct interface *interface)
 {
-	const char *whitespace[] = {
-		"\t\t\t\t\t\t\t\t\t\t\t\t",
-		"\t\t\t\t\t\t\t\t\t\t\t\t ",
-		"\t\t\t\t\t\t\t\t\t\t\t\t  ",
-		"\t\t\t\t\t\t\t\t\t\t\t\t   ",
-		"\t\t\t\t\t\t\t\t\t\t\t\t    ",
-		"\t\t\t\t\t\t\t\t\t\t\t\t     ",
-		"\t\t\t\t\t\t\t\t\t\t\t\t      ",
-		"\t\t\t\t\t\t\t\t\t\t\t\t       "
-	};
+	struct message *m;
+	struct arg *a;
 
-	return whitespace[n % 8] + 12 - n / 8;
+	/* We provide hand written functions for the display object */
+	if (strcmp(interface->name, "wl_display") == 0)
+		return;
+
+	wl_list_for_each(m, message_list, link) {
+		printf("static inline void\n"
+		       "%s_send_%s(struct wl_resource *resource_",
+		       interface->name, m->name);
+
+		wl_list_for_each(a, &m->arg_list, link) {
+			printf(", ");
+			switch (a->type) {
+			case NEW_ID:
+			case OBJECT:
+				printf("struct wl_resource *");
+				break;
+			default:
+				emit_type(a);
+			}
+			printf("%s", a->name);
+		}
+
+		printf(")\n"
+		       "{\n"
+		       "\twl_resource_post_event(resource_, %s_%s",
+		       interface->uppercase_name, m->uppercase_name);
+
+		wl_list_for_each(a, &m->arg_list, link)
+			printf(", %s", a->name);
+
+		printf(");\n");
+		printf("}\n\n");
+	}
 }
 
 static void
@@ -591,20 +738,20 @@ emit_enumerations(struct interface *interface)
 		       interface->uppercase_name, e->uppercase_name);
 
 		if (desc) {
-			printf("/**\n"
-			       " * %s_%s - %s\n", interface->name,
-			       e->name, desc->summary);
+			printf("/**\n");
+			desc_dump(desc->summary,
+				  " * %s_%s - ",
+				  interface->name, e->name);
 			wl_list_for_each(entry, &e->entry_list, link) {
-				printf(" * @%s_%s_%s: %s\n",
-				       interface->uppercase_name,
-				       e->uppercase_name,
-				       entry->uppercase_name,
-				       entry->summary);
+				desc_dump(entry->summary,
+					  " * @%s_%s_%s: ",
+					  interface->uppercase_name,
+					  e->uppercase_name,
+					  entry->uppercase_name);
 			}
 			if (desc->text) {
-				printf(" *\n"
-				       " * ");
-				desc_dump(desc->text, 0);
+				printf(" *\n");
+				desc_dump(desc->text, " * ");
 			}
 			printf(" */\n");
 		}
@@ -633,16 +780,16 @@ emit_structs(struct wl_list *message_list, struct interface *interface)
 	is_interface = message_list == &interface->request_list;
 	if (interface->description) {
 		struct description *desc = interface->description;
-		printf("/**\n"
-		       " * %s - %s\n", interface->name, desc->summary);
+		printf("/**\n");
+		desc_dump(desc->summary, " * %s - ", interface->name);
 		wl_list_for_each(m, message_list, link) {
 			struct description *mdesc = m->description;
-			printf(" * @%s: %s\n", m->name, mdesc ? mdesc->summary :
-			       "(none)");
+			desc_dump(mdesc ? mdesc->summary : "(none)",
+				  " * @%s: ",
+				  m->name);
 		}
-		printf(" *\n"
-		       " * ");
-		desc_dump(desc->text, 0);
+		printf(" *\n");
+		desc_dump(desc->text, " * ");
 		printf(" */\n");
 	}
 	printf("struct %s_%s {\n", interface->name,
@@ -652,16 +799,25 @@ emit_structs(struct wl_list *message_list, struct interface *interface)
 		struct description *mdesc = m->description;
 
 		printf("\t/**\n");
-		printf("\t * %s - %s\n", m->name, mdesc ? mdesc->summary :
-		       "(none)");
+		desc_dump(mdesc ? mdesc->summary : "(none)",
+			  "\t * %s - ", m->name);
 		wl_list_for_each(a, &m->arg_list, link) {
-			printf("\t * @%s: %s\n", a->name, a->summary ?
-			       a->summary : "(none)");
+
+			if (is_interface && a->type == NEW_ID &&
+			    a->interface_name == NULL)
+				printf("\t * @interface: name of the objects interface\n"
+				       "\t * @version: version of the objects interface\n");
+
+
+			desc_dump(a->summary ? a->summary : "(none)",
+				  "\t * @%s: ", a->name);
 		}
 		if (mdesc) {
-			printf("\t * ");
-			desc_dump(mdesc->text, 8);
-			printf("\n");
+			printf("\t *\n");
+			desc_dump(mdesc->text, "\t * ");
+		}
+		if (m->since > 1) {
+			printf("\t * @since: %d\n", m->since);
 		}
 		printf("\t */\n");
 		printf("\tvoid (*%s)(", m->name);
@@ -682,6 +838,13 @@ emit_structs(struct wl_list *message_list, struct interface *interface)
 
 			if (is_interface && a->type == OBJECT)
 				printf("struct wl_resource *");
+			else if (is_interface && a->type == NEW_ID && a->interface_name == NULL)
+				printf("const char *interface, uint32_t version, uint32_t ");
+			else if (!is_interface && a->type == OBJECT && a->interface_name == NULL)
+				printf("struct wl_object *");
+
+			else if (!is_interface && a->type == NEW_ID)
+				printf("struct %s *", a->interface_name);
 			else
 				emit_type(a);
 
@@ -750,11 +913,12 @@ emit_header(struct protocol *protocol, int server)
 	       "\n"
 	       "#include <stdint.h>\n"
 	       "#include <stddef.h>\n"
-	       "#include \"wayland-util.h\"\n\n"
+	       "#include \"%s\"\n\n"
 	       "struct wl_client;\n"
 	       "struct wl_resource;\n\n",
 	       protocol->uppercase_name, s,
-	       protocol->uppercase_name, s);
+	       protocol->uppercase_name, s,
+	       server ? "wayland-util.h" : "wayland-client.h");
 
 	wl_list_for_each(i, &protocol->interface_list, link)
 		printf("struct %s;\n", i->name);
@@ -774,6 +938,7 @@ emit_header(struct protocol *protocol, int server)
 		if (server) {
 			emit_structs(&i->request_list, i);
 			emit_opcodes(&i->event_list, i);
+			emit_event_wrappers(&i->event_list, i);
 		} else {
 			emit_structs(&i->event_list, i);
 			emit_opcodes(&i->request_list, i);
@@ -804,6 +969,9 @@ emit_types_forward_declarations(struct protocol *protocol,
 			switch (a->type) {
 			case NEW_ID:
 			case OBJECT:
+				if (!a->interface_name)
+					continue;
+
 				m->all_null = 0;
 				printf("extern const struct wl_interface %s_interface;\n",
 				       a->interface_name);
@@ -847,8 +1015,7 @@ emit_types(struct protocol *protocol, struct wl_list *message_list)
 			switch (a->type) {
 			case NEW_ID:
 			case OBJECT:
-				if (strcmp(a->interface_name,
-					   "wl_object") != 0)
+				if (a->interface_name)
 					printf("\t&%s_interface,\n",
 					       a->interface_name);
 				else
@@ -879,17 +1046,25 @@ emit_messages(struct wl_list *message_list,
 	wl_list_for_each(m, message_list, link) {
 		printf("\t{ \"%s\", \"", m->name);
 		wl_list_for_each(a, &m->arg_list, link) {
+			if (is_nullable_type(a) && a->nullable)
+				printf("?");
+
 			switch (a->type) {
 			default:
 			case INT:
 				printf("i");
 				break;
 			case NEW_ID:
+				if (a->interface_name == NULL)
+					printf("su");
 				printf("n");
 				break;
 			case UNSIGNED:
 				printf("u");
 				break;
+			case FIXED:
+				printf("f");
+				break;
 			case STRING:
 				printf("s");
 				break;
@@ -947,14 +1122,14 @@ emit_code(struct protocol *protocol)
 		       i->name, i->name, i->version);
 
 		if (!wl_list_empty(&i->request_list))
-			printf("\tARRAY_LENGTH(%s_requests), %s_requests,\n",
-			       i->name, i->name);
+			printf("\t%d, %s_requests,\n",
+			       list_length(&i->request_list), i->name);
 		else
 			printf("\t0, NULL,\n");
 
 		if (!wl_list_empty(&i->event_list))
-			printf("\tARRAY_LENGTH(%s_events), %s_events,\n",
-			       i->name, i->name);
+			printf("\t%d, %s_events,\n",
+			       list_length(&i->event_list), i->name);
 		else
 			printf("\t0, NULL,\n");
 
@@ -976,6 +1151,7 @@ int main(int argc, char *argv[])
 	protocol.type_index = 0;
 	protocol.null_run_length = 0;
 	protocol.copyright = NULL;
+	memset(&ctx, 0, sizeof ctx);
 	ctx.protocol = &protocol;
 
 	ctx.filename = "<stdin>";
@@ -993,7 +1169,7 @@ int main(int argc, char *argv[])
 		buf = XML_GetBuffer(ctx.parser, XML_BUFFER_SIZE);
 		len = fread(buf, 1, XML_BUFFER_SIZE, stdin);
 		if (len < 0) {
-			fprintf(stderr, "fread: %s\n", strerror(errno));
+			fprintf(stderr, "fread: %m\n");
 			exit(EXIT_FAILURE);
 		}
 		XML_ParseBuffer(ctx.parser, len, len == 0);
